import string
import random
import requests
import dataclasses
from json import JSONDecodeError
from .containers.Profile import Profile
from .containers.ProfileHost import ProfileHost
from .lib.Exceptions import APIError, NetworkError
from collections.abc import Generator
from .containers.Post import Post
from .containers.PostUser import PostUser


class Guest:
    request_session: requests.Session = None
    homepage_source: str = None
    insta_app_id: str = "936619743392459"
    preferred_color_scheme: str = "dark"
    x_ig_www_claim: str
    csrf_token: str = None
    user_agent: str = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 " \
                      "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36"

    def __init__(self, proxy: dict[str, str] | None = None) -> None:
        self.request_session = requests.Session()
        self.x_ig_www_claim = "hmac." + "".join(random.choices(string.ascii_letters + string.digits + "_-", k=48))
        self.csrf_token = "".join(random.choices(string.ascii_letters + string.digits, k=32))
        self.request_session.cookies.set("csrftoken", self.csrf_token)

        if proxy is not None: self.request_session.proxies.update(proxy)

    def username_availability(self, username: str) -> bool | None:
        username = username.replace(" ", "").lower()

        body_json = {
            "email": f"{username}@{self.csrf_token}.com",
            "username": username,
            "first_name": username.capitalize(),
            "opt_into_one_tap": False
        }
        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": "0",
            "x-instagram-ajax": "1007614758",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/emailsignup/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.post(
                "https://www.instagram.com/api/v1/web/accounts/web_create_ajax/attempt/",
                headers=request_headers,
                data=body_json
            )
            response_json = http_response.json()

            if "errors" in response_json:
                return "username" not in response_json["errors"]
        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def profile(self, username: str, __session__: requests.Session | None = None) -> Profile | ProfileHost | None:
        username: str = username.replace(" ", "").lower()

        request_headers: dict = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": "0",
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{username}/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            session: requests.Session = __session__
            if __session__ is None: session: requests.Session = self.request_session

            http_response: requests.Response = session.get(
                f"https://www.instagram.com/api/v1/users/web_profile_info/?username={username}",
                headers=request_headers,
            )
            response_json: dict = http_response.json()

            if "status" in response_json:
                if response_json["status"] == "ok" and "data" in response_json:
                    if "user" in response_json["data"]:
                        try:
                            data: dict = response_json["data"]["user"]

                            if data is None:
                                raise APIError("User object not found inside HTTP response.")

                            profile = Profile(
                                raw=data,
                                biography=data["biography"],
                                biography_links=data["bio_links"],
                                country_block=data["country_block"],
                                full_name=data["full_name"],
                                follower_count=data["edge_followed_by"]["count"],
                                following_count=data["edge_follow"]["count"],
                                user_id=data["id"],
                                category_name=data["category_name"],
                                is_business_account=data["is_business_account"],
                                is_professional_account=data["is_professional_account"],
                                is_supervision_enabled=data["is_supervision_enabled"],
                                is_joined_recently=data["is_joined_recently"],
                                is_private=data["is_private"],
                                is_verified=data["is_verified"],
                                profile_picture_url=data["profile_pic_url"],
                                profile_picture_url_hd=data["profile_pic_url_hd"],
                                pronouns=data["pronouns"],
                                has_ar_effects=data["has_ar_effects"],
                                has_clips=data["has_clips"],
                                has_guides=data["has_guides"],
                                has_channel=data["has_channel"],
                                highlight_count=data["highlight_reel_count"],
                                hide_like_and_view_counts=data["hide_like_and_view_counts"],
                                is_embeds_disabled=data["is_embeds_disabled"],
                                is_verified_by_mv4b=data["is_verified_by_mv4b"],
                                should_show_category=data["should_show_category"],
                                should_show_public_contacts=data["should_show_public_contacts"],
                                show_account_transparency_details=data["show_account_transparency_details"],
                                total_post_count=data["edge_owner_to_timeline_media"]["count"]
                            )

                            if __session__ is not None:
                                profile_host = ProfileHost()

                                for key, value in dataclasses.asdict(profile).items():
                                    setattr(profile_host, key, value)

                                profile_host.blocked_by_viewer = data["blocked_by_viewer"]
                                profile_host.followed_by_viewer = data["followed_by_viewer"]
                                profile_host.follows_viewer = data["follows_viewer"]
                                profile_host.has_blocked_viewer = data["has_blocked_viewer"]
                                profile_host.has_requested_viewer = data["has_requested_viewer"]
                                profile_host.is_guardian_of_viewer = data["is_guardian_of_viewer"]
                                profile_host.is_supervised_by_viewer = data["is_supervised_by_viewer"]
                                profile_host.requested_by_viewer = data["requested_by_viewer"]
                                profile_host.mutual_follower_count = data["edge_mutual_followed_by"]["count"]

                                return profile_host

                            return profile
                        except KeyError:
                            raise APIError()
        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def get_uid(self, username: str, __session__: requests.Session | None = None) -> str | None:
        username: str = username.strip().lower().replace(" ", "")
        response: Profile | None = self.profile(username, __session__)

        if response.user_id is not None:
            return response.user_id.replace(" ", "")

    def get_username(self, uid: str | int, __session__: requests.Session | None = None) -> str | None:
        uid = uid.replace(" ", "")

        request_headers = {
            "User-Agent": "Instagram 76.0.0.15.395 Android (24/7.0; 640dpi; 1440x2560; "
                          "samsung; SM-G930F; herolte; samsungexynos8890; en_US; 138226743)"
        }

        try:
            session: requests.Session = __session__
            if __session__ is None: session: requests.Session = self.request_session

            http_response = session.get(f"https://i.instagram.com/api/v1/users/{uid}/info/", headers=request_headers)
            response_json = http_response.json()

            if "status" in response_json \
                    and response_json["status"] == "ok" \
                    and "user" in response_json \
                    and "username" in response_json["user"]:

                return response_json["user"]["username"].replace(" ", "").lower()

        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def posts(self, username: str, count: int = 0, __session__: requests.Session | None = None) -> Generator[Post, None, None]:
        """
        Generates a list of target's posts of specified size.
        :param username: Target's Username
        :param count: Amount of posts to fetch
        :param __session__: (Optional) Custom request session object
        :return: Generator which yields each post's data
        """

        username = username.replace(" ", "").lower()

        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{username}/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        current_max_id = ""
        generated_count = 0

        while True:
            current_max_id_text = ""

            if current_max_id != "":
                current_max_id_text = f"&max_id={current_max_id}"

            try:
                count_text = 35

                if count < 35:
                    count_text = count

                session: requests.Session = __session__
                if __session__ is None: session: requests.Session = self.request_session

                http_response = session.get(
                    f"https://www.instagram.com/api/v1/feed/user/{username}/username/?count={count_text}"
                    f"{current_max_id_text}",
                    headers=request_headers
                )
                response_json = http_response.json()

                if "status" not in response_json or "items" not in response_json:
                    yield None
                    raise NetworkError("HTTP response doesn't include 'status' or 'items' node.")

                if response_json["status"] != "ok":
                    yield None
                    raise NetworkError("HTTP response status not 'ok'.")

                for each_item in response_json["items"]:
                    if generated_count < count or count == 0:

                        yield self.__process_post_data(each_item)
                        generated_count += 1

                if (generated_count < count or count == 0) and "next_max_id" in response_json:
                    current_max_id = response_json["next_max_id"]
                else:
                    return None
            except JSONDecodeError:
                yield None
                raise NetworkError("HTTP Response is not a valid JSON.")

    @staticmethod
    def __process_post_data(data: dict) -> Post:

        caption: dict = data.get("caption", None)

        caption_text = ""
        is_caption_covered = False
        caption_created_at = 0
        caption_share_enabled = False
        caption_did_report_as_spam = False

        if caption is not None:
            caption_text = caption.get("text", "")
            is_caption_covered = caption.get("is_covered", False)
            caption_created_at = caption.get("created_at", 0)
            caption_share_enabled = caption.get("share_enabled", False)
            caption_did_report_as_spam = caption.get("did_report_as_spam", False)

        user: PostUser = PostUser()
        user_data: dict = data.get("user", None)

        if user_data is not None:
            user: PostUser = PostUser(
                has_anonymous_profile_picture=user_data.get("has_anonymous_profile_picture", False),
                fbid_v2=user_data.get("fbid_v2", ""),
                transparency_product_enabled=user_data.get("transparency_product_enabled", False),
                is_favorite=user_data.get("is_favorite", False),
                is_unpublished=user_data.get("is_unpublished", False),
                uid=user_data.get("pk", ""),
                username=user_data.get("username", ""),
                full_name=user_data.get("full_name", ""),
                is_private=user_data.get("is_private", False),
                is_verified=user_data.get("is_verified", False),
                profile_picture_id=user_data.get("profile_pic_id", ""),
                profile_picture_url=user_data.get("profile_pic_url", ""),
                account_badges=user_data.get("account_badges", []),
                feed_post_reshare_disabled=user_data.get("feed_post_reshare_disabled", False),
                show_account_transparency_details=user_data.get("show_account_transparency_details", False),
                third_party_downloads_enabled=user_data.get("third_party_downloads_enabled", 0),
                latest_reel_media=user_data.get("latest_reel_media", 0)
            )

        return Post(
            share_url=f"https://www.instagram.com/p/{data.get('code', '')}",
            taken_at=data.get("taken_at", 0),
            post_id=data.get("pk", ""),
            media_type=data.get("media_type", 0),
            code=data.get("code", ""),
            caption_is_edited=data.get("caption_is_edited", False),
            original_media_has_visual_reply_media=data.get("original_media_has_visual_reply_media", False),
            like_and_view_counts_disabled=data.get("like_and_view_counts_disabled", False),
            can_viewer_save=data.get("can_viewer_save", False),
            profile_grid_control_enabled=data.get("profile_grid_control_enabled", False),
            is_comments_gif_composer_enabled=data.get("is_comments_gif_composer_enabled", False),
            comment_threading_enabled=data.get("comment_threading_enabled", False),
            comment_count=data.get("comment_count", 0),
            has_liked=data.get("has_liked", False),
            user=user,
            can_viewer_reshare=data.get("can_viewer_reshare", False),
            like_count=data.get("like_count", 0),
            top_likers=data.get("top_likers", []),
            caption_text=caption_text,
            is_caption_covered=is_caption_covered,
            caption_created_at=caption_created_at,
            caption_share_enabled=caption_share_enabled,
            caption_did_report_as_spam=caption_did_report_as_spam,
            is_paid_partnership=data.get("is_paid_partnership", False),
            show_shop_entrypoint=data.get("show_shop_entrypoint", False),
            deleted_reason=data.get("deleted_reason", 0),
            integrity_review_decision=data.get("integrity_review_decision", ""),
            ig_media_sharing_disabled=data.get("ig_media_sharing_disabled", False),
            has_shared_to_fb=data.get("has_shared_to_fb", False),
            is_unified_video=data.get("is_unified_video", False),
            should_request_ads=data.get("should_request_ads", False),
            is_visual_reply_commenter_notice_enabled=data.get("is_visual_reply_commenter_notice_enabled", False),
            commerciality_status=data.get("commerciality_status", ""),
            explore_hide_comments=data.get("explore_hide_comments", False),
            has_delayed_metadata=data.get("has_delayed_metadata", False),
            location_latitude=data.get("lat", 0),
            location_longitude=data.get("lng", 0)
        )



import time
import json
import random
import string
import requests
import moviepy.editor
from uuid import uuid4
from .Guest import Guest
from pathlib import Path
from .Direct import Direct
from json import JSONDecodeError
from .containers.Liker import Liker
from .containers.Likers import Likers
from .containers.Post import Post
from collections.abc import Generator
from .containers.ProfileHost import ProfileHost
from .containers.PrivateInfo import PrivateInfo
from .containers import (FollowedStatus, UnfollowedStatus, FollowPerson)
from .lib import (
    SessionError,
    NetworkError,
    IdentifierError,
    DevelopmentError,
    APIError,
    ConversionError,
    FileTypeError
)

USERNAME, UID = 0, 1


class SessionHost:
    session_data: str
    request_session: requests.Session
    insta_app_id: str = "936619743392459"
    preferred_color_scheme: str = "dark"
    x_ig_www_claim: str
    csrf_token: str
    guest: Guest
    user_id: str
    user_agent: str = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 " \
                      "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36"

    def __init__(
        self,
        session_data: str,
        proxy: dict[str, str] = None,
        skip_auth_verification: bool = False
    ) -> None:

        self.session_data = session_data
        self.x_ig_www_claim = "hmac." + "".join(random.choices(string.ascii_letters + string.digits + "_-", k=48))
        self.csrf_token = "".join(random.choices(string.ascii_letters + string.digits, k=32))
        self.request_session = requests.Session()

        if proxy is not None: self.request_session.proxies.update(proxy)

        self.guest = Guest(proxy=proxy)

        session_data_json: dict = json.loads(session_data)
        self.user_id = session_data_json.get("user_id")

        self.request_session.cookies.set("sessionid", session_data_json.get("session_id"))
        self.request_session.cookies.set("rur", session_data_json.get("rur"))
        self.request_session.cookies.set("mid", session_data_json.get("mid"))
        self.request_session.cookies.set("ds_user_id", session_data_json.get("user_id"))
        self.request_session.cookies.set("ig_did", session_data_json.get("ig_did"))
        self.request_session.cookies.set("csrftoken", self.csrf_token)

        if not skip_auth_verification and not self.authenticated():
            raise SessionError(
                "SessionID expired. If you used a saved session, delete ensta-session.txt file and try again"
            )

    def authenticated(self) -> bool:
        """
        Returns whether user is logged in or not.
        That is, if the provided username-password combination or session-id is correct.
        :return: Boolean (True / False)
        """

        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/edit/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }
        http_response = self.request_session.get(
            "https://www.instagram.com/api/v1/accounts/edit/web_form_data/",
            headers=request_headers
        )

        try:
            http_response.json()
            return True
        except JSONDecodeError:
            return False

    def follow(self, identifier: str | int) -> FollowedStatus | None:
        """
        Follows the target user.
        :param identifier: Target user's Username or UserID
        :return: Object with the followed info
        """

        conversion_success, identifier = self._identifier(identifier, UID)
        if not conversion_success: raise ConversionError(f"Can't convert identifier \"{identifier}\" into 'UID'.")

        # Actual Request
        body_json = {
            "container_module": "profile",
            "nav_chain": f"PolarisProfileNestedContentRoot:profilePage:1:via_cold_start",
            "user_id": identifier
        }
        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1007616494",
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{''.join(random.choices(string.ascii_lowercase, k=5))}/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.post(
                f"https://www.instagram.com/api/v1/friendships/create/{identifier}/",
                headers=request_headers,
                data=body_json
            )
            response_json = http_response.json()

            if "status" in response_json:
                if response_json["status"] == "ok" and "friendship_status" in response_json:
                    if "following" in response_json["friendship_status"] \
                            and "outgoing_request" in response_json["friendship_status"] \
                            and "followed_by" in response_json["friendship_status"] \
                            and "previous_following" in response_json:
                        return FollowedStatus(
                            following=response_json["friendship_status"]["following"],
                            follow_requested=response_json["friendship_status"]["outgoing_request"],
                            is_my_follower=response_json["friendship_status"]["followed_by"],
                            previous_following=response_json["previous_following"]
                        )
                else:
                    if response_json["status"] != "ok": raise NetworkError("Response \"Status\" not ok.")
                    else: raise NetworkError("'friendship_status' attribute not in response json.")
            else:
                raise NetworkError("'status' attribute not in response json.")
        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def unfollow(self, identifier: str | int) -> UnfollowedStatus | None:
        """
        Unfollows the target user.
        :param identifier: Target user's Username or UserID
        :return: Object with the unfollowed info
        """

        conversion_success, identifier = self._identifier(identifier, UID)
        if not conversion_success: raise ConversionError(f"Can't convert identifier \"{identifier}\" into 'UID'.")

        # Actual Request
        body_json = {
            "container_module": "profile",
            "nav_chain": f"PolarisProfileRoot:profilePage:1:via_cold_start",
            "user_id": identifier
        }
        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1007616494",
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{''.join(random.choices(string.ascii_lowercase, k=6))}/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.post(
                f"https://www.instagram.com/api/v1/friendships/destroy/{identifier}/",
                headers=request_headers,
                data=body_json
            )
            response_json = http_response.json()

            if "status" in response_json:
                if response_json["status"] == "ok" and "friendship_status" in response_json:
                    if "following" in response_json["friendship_status"] \
                            and "outgoing_request" in response_json["friendship_status"] \
                            and "followed_by" in response_json["friendship_status"]:

                        return UnfollowedStatus(
                            unfollowed=not (
                                    response_json["friendship_status"]["following"] or
                                    response_json["friendship_status"]["outgoing_request"]
                            ),

                            is_my_follower=response_json["friendship_status"]["followed_by"]
                        )
                else:
                    if response_json["status"] != "ok": raise NetworkError("Response \"Status\" not ok.")
                    else: raise NetworkError("'friendship_status' attribute not in response json.")
            else:
                raise NetworkError("'status' attribute not in response json.")
        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def followers(self, identifier: str | int, count: int = 0) -> Generator[FollowPerson, None, None]:
        """
        Generates a list of target's followers of specified size.

        :param identifier: Target's Username or UserID
        :param count: Amount of followers to fetch
        :return: Generator which yields each user's details
        """

        conversion_success, identifier = self._identifier(identifier, UID)
        if not conversion_success:
            yield None
            raise ConversionError(f"Can't convert identifier \"{identifier}\" into 'UID'.")

        # Actual Request
        request_headers: dict = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{''.join(random.choices(string.ascii_lowercase, k=6))}/followers/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        current_max_id: str = ""
        generated_count: int = 0

        while True:
            current_max_id_text: str = ""

            if current_max_id != "":
                current_max_id_text: str = f"&max_id={current_max_id}"

            try:
                count_text = 35

                if count < 35:
                    count_text = count

                http_response = self.request_session.get(
                    f"https://www.instagram.com/api/v1/friendships/{identifier}/followers/?count={str(count_text)}"
                    f"{current_max_id_text}&search_surface=follow_list_page",
                    headers=request_headers
                )
                response_json = http_response.json()

                if "status" not in response_json or "users" not in response_json:
                    yield None
                    raise NetworkError("HTTP response doesn't include 'status' or 'users' node.")

                if response_json["status"] != "ok":
                    yield None
                    raise NetworkError("HTTP response status not 'ok'.")

                for each_item in response_json["users"]:
                    if generated_count < count or count == 0:

                        try:
                            yield FollowPerson(
                                has_anonymous_profile_picture=each_item["has_anonymous_profile_picture"],
                                user_id=each_item["pk"],
                                username=each_item["username"],
                                full_name=each_item["full_name"],
                                is_private=each_item["is_private"],
                                is_verified=each_item["is_verified"],
                                profile_picture_url=each_item["profile_pic_url"],
                                badges=each_item["account_badges"],
                                third_party_downloads_enabled=each_item["third_party_downloads_enabled"]
                            )

                        except KeyError:
                            raise APIError()

                        generated_count += 1

                if (generated_count < count or count == 0) and "next_max_id" in response_json:
                    current_max_id = response_json["next_max_id"]
                else:
                    return None
            except JSONDecodeError:
                yield None
                raise NetworkError("HTTP Response is not a valid JSON.")

    def followings(self, identifier: str | int, count: int = 0) -> Generator[FollowPerson, None, None]:
        """
        Generates a list of users which the target follows, of specified size.
        :param identifier: Target's Username or UserID
        :param count: Amount of followings to fetch
        :return: Generator which yields each user's details
        """

        conversion_success, identifier = self._identifier(identifier, UID)
        if not conversion_success:
            yield None
            raise ConversionError(f"Can't convert identifier \"{identifier}\" into 'UID'.")

        # Actual Request
        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/{''.join(random.choices(string.ascii_lowercase, k=6))}/following/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        current_max_id = ""
        generated_count = 0

        while True:
            current_max_id_text = ""

            if current_max_id != "":
                current_max_id_text = f"&max_id={current_max_id}"

            try:
                count_text = 35

                if count < 35:
                    count_text = count

                http_response = self.request_session.get(
                    f"https://www.instagram.com/api/v1/friendships/{identifier}/following/?count={str(count_text)}"
                    f"{current_max_id_text}",
                    headers=request_headers)
                response_json = http_response.json()

                if "status" not in response_json or "users" not in response_json:
                    yield None
                    raise NetworkError("HTTP response doesn't include 'status' or 'users' node.")

                if response_json["status"] != "ok":
                    yield None
                    raise NetworkError("HTTP response status not 'ok'.")

                for each_item in response_json["users"]:
                    if generated_count < count or count == 0:

                        try:
                            yield FollowPerson(
                                has_anonymous_profile_picture=each_item["has_anonymous_profile_picture"],
                                user_id=each_item["pk"],
                                username=each_item["username"],
                                full_name=each_item["full_name"],
                                is_private=each_item["is_private"],
                                is_verified=each_item["is_verified"],
                                profile_picture_url=each_item["profile_pic_url"]
                            )

                        except KeyError:
                            raise APIError()

                        generated_count += 1

                if (generated_count < count or count == 0) and "next_max_id" in response_json:
                    current_max_id = response_json["next_max_id"]
                else:
                    return None
            except JSONDecodeError:
                yield None
                raise NetworkError("HTTP Response is not a valid JSON.")

    def _identifier(self, identifier: str | int, required: str | int):
        identifier = str(identifier).lower().replace(" ", "")

        if len(identifier) <= 0:
            raise IdentifierError("No identifier was given. Please pass either UserId or Username as an argument.")

        # Identifier: Username or UID?
        is_username = False
        for letter in identifier:
            if letter not in string.digits:
                is_username = True
                break

        if is_username and required == USERNAME:
            return True, identifier

        elif is_username and required == UID:
            user_id = self.get_uid(identifier)

            if user_id is not None and user_id != "":
                return True, user_id
            else:
                return False, None

        elif not is_username and required == USERNAME:
            username = self.get_username(identifier)

            if username is not None and username != "":
                return True, username
            else:
                return False, None

        elif not is_username and required == UID:
            return True, identifier

        else:
            raise DevelopmentError()

    def _set_account_privacy(self, privacy: str) -> bool:
        is_private = (privacy == "private")

        if privacy != "private" and privacy != "public":
            raise DevelopmentError()

        body_json = {
            "is_private": is_private
        }
        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "198387",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1007616494",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/who_can_see_your_content/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.post(
                "https://www.instagram.com/api/v1/web/accounts/set_private/",
                headers=request_headers,
                data=body_json
            )
            response_json = http_response.json()

            if "status" not in response_json:
                return False

            if response_json["status"] != "ok":
                return False

            return True
        except JSONDecodeError:
            return False

    def switch_to_private_account(self) -> bool:
        """
        Switches your account type to 'Private'. This means, only your followers can see your posts, stories & highlights.
        :return: Boolean (Whether your account type was successfully switched to private)
        """

        return self._set_account_privacy("private")

    def switch_to_public_account(self) -> bool:
        """
        Switches your account type to 'Public'. This means, anyone can see your posts, stories & highlights.
        :return: Boolean (Whether your account type was successfully switched to public)
        """

        return self._set_account_privacy("public")

    def profile(self, username: str) -> ProfileHost | None:
        """
        Returns profile data of the target user. Including name, profile picture, follower_count, following_count, etc.
        :param username: Username of the target
        :return: Profile object which contains all the data
        """

        return self.guest.profile(username, __session__=self.request_session)

    def get_username(self, uid: str | int) -> str | None:
        """
        Returns the username of the target user, when given their userid.
        :param uid: Target's UserID
        :return: Target's Username
        """

        return self.guest.get_username(uid, __session__=self.request_session)

    def get_uid(self, username: str) -> str | None:
        """
        Returns the userid of the target user, when given their username.
        :param username: Target's Username
        :return: Target's UserID
        """

        return self.guest.get_uid(username, __session__=self.request_session)

    def posts(self, username: str, count: int = 0) -> Generator[Post, None, None]:
        """
        Generates a list of target's posts of specified size.
        :param username: Target's Username
        :param count: Amount of posts to fetch
        :return: Generator which yields each post's data
        """

        return self.guest.posts(username, count, __session__=self.request_session)

    def get_post_id(self, share_url: str) -> str:
        """
        Returns post_id of specific post, given its share_url, which can further be used to like or add comment to that post.
        :param share_url: Share URL of post. e.g. - https://www.instagram.com/p/Czr2yLmroCQ/
        :return: PostID in text format
        """

        share_url: str = share_url.strip()

        request_headers = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif"
                      ",image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-language": "en-US,en;q=0.9",
            "cache-control": "max-age=0",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "same-origin",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "viewport-width": "1475",
            "referrerPolicy": "strict-origin-when-cross-origin"
        }

        http_response = self.request_session.get(share_url, headers=request_headers)
        response_text = http_response.text

        required_text = "instagram://media?id="

        initial_index = response_text.find(required_text)
        if initial_index == -1: raise APIError()

        rest_text = response_text[initial_index + len(required_text): initial_index + len(required_text) + 25]

        end_index = rest_text.find('"')
        if end_index == -1: raise APIError()

        return rest_text[:end_index]

    def private_info(self) -> PrivateInfo:
        """
        Returns your own (logged-in user's) private info which only you can access. Such as gender, birthday, email, phone number, etc.
        :return: Object which contains your private info
        """

        request_headers = {
            "accept": "*/*",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/edit/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        http_response = self.request_session.get(
            f"https://www.instagram.com/api/v1/accounts/edit/web_form_data/",
            headers=request_headers
        )

        try:
            response_json: dict = http_response.json()

            if "status" not in response_json: raise NetworkError("Key 'status' not in response json.")
            if response_json.get("status") != "ok": raise NetworkError("Key 'status' not 'ok' in response json.")

            data: json = response_json.get("form_data", {})
            if len(data.keys()) == 0: raise NetworkError("Form data doesn't contain any keys.")

            return PrivateInfo(
                first_name=data.get("first_name", ""),
                last_name=data.get("last_name", ""),
                email=data.get("email", ""),
                is_email_confirmed=data.get("is_email_confirmed", False),
                is_phone_confirmed=data.get("is_phone_confirmed", False),
                username=data.get("username", ""),
                phone_number=data.get("phone_number", ""),
                gender=("f" if data.get("gender", 0) == 0 else "m"),
                birthday=data.get("birthday", ""),
                fb_birthday=data.get("fb_birthday", ""),
                biography=data.get("biography", ""),
                external_url=data.get("external_url", ""),
                chaining_enabled=data.get("chaining_enabled", False),
                presence_disabled=data.get("presence_disabled", False),
                business_account=data.get("business_account", False),
                user_tag_review_enabled=data.get("usertag_review_enabled", False),
                custom_gender=data.get("custom_gender", ""),
                trusted_username=data.get("trusted_username", ""),
                trust_days=data.get("trust_days", 0)
            )

        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def change_bio(self, biography: str) -> bool:
        """
        Updates your (logged-in user's) biography to the newly given text.
        :param biography: New Biography
        :return: Boolean (Whether your biography was successfully updated or not)
        """

        private_info = self.private_info()

        request_headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1009841815",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/edit/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json = {
            "biography": biography,
            "chaining_enabled": "on" if private_info.chaining_enabled else "off",
            "email": private_info.email,
            "external_url": private_info.external_url,
            "first_name": private_info.first_name,
            "phone_number": private_info.phone_number,
            "username": private_info.username
        }

        http_response = self.request_session.post(
            f"https://www.instagram.com/api/v1/web/accounts/edit/",
            headers=request_headers,
            data=body_json
        )

        try:
            response_json: dict = http_response.json()

            if "status" not in response_json: raise NetworkError(
                "Key 'status' not in response json. Possibly it's a fault from your side."
            )
            return response_json.get("status") == "ok"

        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def change_display_name(self, display_name: str) -> bool:
        """
        Updates your (logged-in user's) display name on Instagram.
        :param display_name: New name
        :return: Boolean (Whether your name was successfully updated or not)
        """

        private_info = self.private_info()

        request_headers = {
            "accept": "*/*",
            "content-type": "application/x-www-form-urlencoded",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1009841815",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/accounts/edit/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json = {
            "biography": private_info.biography,
            "chaining_enabled": "on" if private_info.chaining_enabled else "off",
            "email": private_info.email,
            "external_url": private_info.external_url,
            "first_name": display_name,
            "phone_number": private_info.phone_number,
            "username": private_info.username
        }

        http_response = self.request_session.post(
            f"https://www.instagram.com/api/v1/web/accounts/edit/",
            headers=request_headers,
            data=body_json
        )

        try:
            response_json: dict = http_response.json()

            if "status" not in response_json: raise NetworkError(
                "Key 'status' not in response json. Possibly it's a fault from your side."
            )

            return response_json.get("status") == "ok"

        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    def get_upload_id(self, media_path: str, arg_upload_id: str | None = None) -> str:
        """
        Uploads the given media to Instagram's server and returns its unique ID which you can later use to configure single or multiple posts.
        :param media_path: Path to the media file (only jpg & jpeg)
        :param arg_upload_id: Custom upload_id (for advanced users)
        :return: Upload ID of uploaded file
        """

        media_path: Path = Path(media_path)

        if media_path.suffix not in (".jpg", ".jpeg"): raise FileTypeError(
            "Only jpg and jpeg image types are allowed to post."
        )

        upload_id = arg_upload_id if arg_upload_id is not None else str(int(time.time()) * 1000)
        waterfall_id = str(uuid4())
        upload_name = f"{upload_id}_0_{random.randint(1000000000, 9999999999)}"

        rupload_params = {
            "retry_context": "{\"num_step_auto_retry\": 0, \"num_reupload\": 0, \"num_step_manual_retry\": 0}",
            "media_type": "1",
            "xsharing_user_ids": "[]",
            "upload_id": upload_id,
            "image_compression": json.dumps({"lib_name": "moz", "lib_version": "3.1.m", "quality": 80})
        }

        with open(media_path, "rb") as file:
            photo_data = file.read()
            photo_length = str(len(photo_data))

        request_headers = {
            "accept-encoding": "gzip",
            "x-instagram-rupload-params": json.dumps(rupload_params),
            "x_fb_photo_waterfall_id": waterfall_id,
            "x-entity-type": "image/jpeg",
            "offset": "0",
            "x-entity-name": upload_name,
            "x-entity-length": photo_length,
            "content-type": "application/octet-stream",
            "content-length": photo_length
        }

        http_response = self.request_session.post(
            f"https://i.instagram.com/rupload_igphoto/{upload_name}",
            data=photo_data,
            headers=request_headers
        )

        try:
            response_json: dict = http_response.json()

            if response_json.get("status", "") != "ok": raise NetworkError("Response json key 'status' not ok.")
            if response_json.get("upload_id", "") == "": raise NetworkError(
                "Key 'upload_id' in response json doesn't exist or is invalid."
            )

            return str(response_json.get("upload_id"))

        except JSONDecodeError:
            raise NetworkError("Response not a valid json.")

    def __upload_video(self, path: str, arg_upload_id: str | None = None) -> tuple[bool, any, any, any]:
        video_editor = moviepy.editor.VideoFileClip(path)

        path: Path = Path(path)
        waterfall_id = str(uuid4())

        upload_id = arg_upload_id if arg_upload_id is not None else str(int(time.time()) * 1000)
        upload_name = f"{upload_id}_0_{random.randint(1000000000, 9999999999)}"

        rupload_params = {
            "is_clips_video": "1",
            "retry_context": "{\"num_step_auto_retry\": 0, \"num_reupload\": 0, \"num_step_manual_retry\": 0}",
            "media_type": "2",
            "xsharing_user_ids": json.dumps([self.user_id]),
            "upload_id": upload_id,
            "upload_media_duration_ms": str(int(video_editor.duration * 1000)),
            "upload_media_width": str(video_editor.size[0]),
            "upload_media_height": str(video_editor.size[1])
        }

        request_headers__get = {
            "accept-encoding": "gzip",
            "x-instagram-rupload-params": json.dumps(rupload_params),
            "x_fb_video_waterfall_id": waterfall_id,
            "x-entity-type": "video/mp4"
        }

        http_response__get = self.request_session.get(
            f"https://i.instagram.com/rupload_igvideo/{upload_name}",
            headers=request_headers__get
        )

        if http_response__get.status_code != 200: raise NetworkError(
            "Video Upload 'GET' Request failed. Status code not 200."
        )

        # POST Request

        with open(path, "rb") as file:
            video_data = file.read()
            video_length = str(len(video_data))

        request_headers = {
            "offset": "0",
            "x-entity-name": upload_name,
            "x-entity-length": video_length,
            "content-type": "application/octet-stream",
            "content-length": video_length,
            **request_headers__get
        }

        http_response = self.request_session.post(
            f"https://i.instagram.com/rupload_igvideo/{upload_name}",
            data=video_data,
            headers=request_headers
        )

        try:
            response_json: dict = http_response.json()

            return response_json.get("status", "") == "ok",\
                video_editor.duration,\
                video_editor.size[0],\
                video_editor.size[1]

        except JSONDecodeError:
            raise NetworkError("Response not a valid json.")

    def upload_photo(
        self,
        upload_id: str,
        caption: str = "",
        alt_text: str = "",
        archive_only: bool = False,
        disable_comments: bool = False,
        like_and_view_counts_disabled: bool = False
    ) -> bool:  # TODO: Implement Return Value

        """
        Creates a single photo post on your account.
        :param upload_id: Upload ID of file already uploaded using get_upload_id() method
        :param caption: Optional caption text for current post
        :param alt_text: Optional custom accessibility caption for this photo
        :param archive_only: Boolean (Should this post be directly archived)
        :param disable_comments: Boolean (Should comments on this post be disabled)
        :param like_and_view_counts_disabled: Boolean (Shouldn't people be able to see how many users liked & viewed this post)
        :return: Boolean (Whether post was successfully created or not)
        """

        request_headers: dict = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1009848613",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json = {
            "archive_only": archive_only,
            "caption": caption,
            "clips_share_preview_to_feed": "1",
            "disable_comments": "1" if disable_comments else "0",
            "disable_oa_reuse": False,
            "igtv_share_preview_to_feed": "1",
            "is_meta_only_post": "0",
            "is_unified_video": "1",
            "like_and_view_counts_disabled": "1" if like_and_view_counts_disabled else "0",
            "source_type": "library",
            "upload_id": upload_id,
            "video_subtitles_enabled": "0"
        }

        if alt_text != "": body_json["custom_accessibility_caption"] = alt_text

        http_response = self.request_session.post(
            "https://www.instagram.com/api/v1/media/configure/",
            headers=request_headers,
            data=body_json
        )

        try:
            response_json: dict = http_response.json()

            return response_json.get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError("Response not a valid json.")

    def upload_photos(
        self,
        upload_ids: list[str],
        caption: str = "",
        archive_only: bool = False,
        disable_comments: bool = False,
        like_and_view_counts_disabled: bool = False,
    ) -> bool:  # TODO: Implement Return Value

        """
        Creates a single post with multiple photos on your account.
        :param upload_ids: List (Upload IDs of files already uploaded using get_upload_id() method)
        :param caption: Optional caption text for current post
        :param archive_only: Boolean (Should this post be directly archived)
        :param disable_comments: Boolean (Should comments on this post be disabled)
        :param like_and_view_counts_disabled: Boolean (Shouldn't people be able to see how many users liked & viewed this post)
        :return: Boolean (Whether post was successfully created or not)
        """

        request_headers: dict = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1009848613",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json = {
            "archive_only": archive_only,
            "caption": caption,
            "children_metadata": [{"upload_id": upload_id} for upload_id in upload_ids],
            "client_sidecar_id": str(int(time.time()) * 1000),
            "disable_comments": "1" if disable_comments else "0",
            "like_and_view_counts_disabled": "1" if like_and_view_counts_disabled else "0",
            "source_type": "library"
        }

        http_response = self.request_session.post(
            "https://www.instagram.com/api/v1/media/configure_sidecar/",
            headers=request_headers,
            data=json.dumps(body_json)
        )

        try:
            response_json: dict = http_response.json()

            return response_json.get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError("Response not a valid json.")

    def upload_reel(
        self,
        video_path: str,
        thumbnail_path: str,
        caption: str = "",
        alt_text: str = "",
        archive_only: bool = False,
        disable_comments: bool = False,
        like_and_view_counts_disabled: bool = False,
        video_subtitles_enabled: bool = False
    ) -> bool:  # TODO: Implement Return Value

        """
        Uploads a reel on your account.
        :param video_path: Path of the video file
        :param thumbnail_path: Path of the image to be used as the thumbnail of current reel
        :param caption: Optional caption text for current reel
        :param alt_text: Optional custom accessibility caption for this reel
        :param archive_only: Boolean (Should this reel be directly archived)
        :param disable_comments: Boolean (Should comments on this reel be disabled)
        :param like_and_view_counts_disabled: Boolean (Shouldn't people be able to see how many users liked & viewed this reel)
        :param video_subtitles_enabled: Boolean (Should subtitles be enabled on this reel)
        :return: Boolean (Whether reel was successfully uploaded or not)
        """

        upload_id = str(int(time.time()) * 1000)

        video_success, video_duration, video_width, video_height = self.__upload_video(video_path, upload_id)

        if not video_success: return False
        if not self.get_upload_id(
                media_path=thumbnail_path,
                arg_upload_id=upload_id
        ):
            return False

        request_headers: dict = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "dpr": "1.30208",
            "sec-ch-prefers-color-scheme": "dark",
            "sec-ch-ua": self.user_agent,
            "sec-ch-ua-full-version-list": self.user_agent,
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-model": "\"\"",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1009848613",
            "x-requested-with": "XMLHttpRequest",
            "Referer": "https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json = {
            "archive_only": archive_only,
            "caption": caption,
            "clips_share_preview_to_feed": "1",
            "disable_comments": "1" if disable_comments else "0",
            "disable_oa_reuse": "0",
            "igtv_share_preview_to_feed": "1",
            "is_meta_only_post": "0",
            "is_unified_video": "1",
            "like_and_view_counts_disabled": "1" if like_and_view_counts_disabled else "0",
            "source_type": "library",
            "upload_id": upload_id,
            "video_subtitles_enabled": "1" if video_subtitles_enabled else "0"
        }

        if alt_text != "": body_json["custom_accessibility_caption"] = alt_text

        http_response = self.request_session.post(
            "https://www.instagram.com/api/v1/media/configure_to_clips/",
            headers=request_headers,
            data=body_json
        )

        try:
            response_json: dict = http_response.json()

            return response_json.get("status", "") == "ok"

        except JSONDecodeError:
            raise NetworkError("Response not a valid json.")

    def comment(self, text: str, post_id: str) -> bool:
        """
        Adds a comment on target post.
        :param text: Comment text
        :param post_id: ID of target post, fetch using get_post_id() method
        :return: Boolean (Whether comment was successfully added or not)
        """

        request_headers: json = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": "\"Not.A/Brand\";v=\"8\", \"Chromium\";v=\"114\", \"Google Chrome\";v=\"114\"",
            "sec-ch-ua-full-version-list": "\"Not.A/Brand\";v=\"8.0.0.0\", \"Chromium\";v=\"114.0.5735.134\", "
                                           "\"Google Chrome\";v=\"114.0.5735.134\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        body_json: json = {"comment_text": text}

        http_response = self.request_session.post(
            f"https://www.instagram.com/api/v1/web/comments/{post_id}/add/",
            headers=request_headers,
            data=body_json
        )

        try:
            response_json: dict = http_response.json()

            return response_json.get("status", "") == "ok"

        except JSONDecodeError:
            return False

    def __like_action(self, post_id: str, action: str = "like") -> bool:

        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/x-www-form-urlencoded",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": "\"Not.A/Brand\";v=\"8\", \"Chromium\";v=\"114\", \"Google Chrome\";v=\"114\"",
            "sec-ch-ua-full-version-list": "\"Not.A/Brand\";v=\"8.0.0.0\", \"Chromium\";v=\"114.0.5735.110\", "
                                           "\"Google Chrome\";v=\"114.0.5735.110\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-instagram-ajax": "1007670408",
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.post(
                f"https://www.instagram.com/api/v1/web/likes/{post_id}/{action}/",
                headers=request_headers
            )

            response_json = http_response.json()

            if "status" in response_json:
                return response_json["status"] == "ok"
            else:
                return False
        except JSONDecodeError:
            return False

    def like(self, post_id: str) -> bool:
        """
        Likes the target post.
        :param post_id: ID of target post, fetch using get_post_id() method
        :return: Boolean (Whether target post was liked successfully or not)
        """

        return self.__like_action(post_id, "like")

    def unlike(self, post_id: str) -> bool:
        """
        Unlikes the target post.
        :param post_id: ID of target post, fetch using get_post_id() method
        :return: Boolean (Whether target post was unliked successfully or not)
        """

        return self.__like_action(post_id, "unlike")

    def likers(self, post_id: str) -> Likers | None:
        """
        Generates a list of users who liked the target post of specified size.
        :param post_id: ID of target post, fetch using get_post_id() method
        :return: Generator which yields each user's data
        """

        request_headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9",
            "sec-ch-prefers-color-scheme": self.preferred_color_scheme,
            "sec-ch-ua": "\"Not.A/Brand\";v=\"8\", \"Chromium\";v=\"114\", \"Google Chrome\";v=\"114\"",
            "sec-ch-ua-full-version-list": "\"Not.A/Brand\";v=\"8.0.0.0\", \"Chromium\";v=\"114.0.5735.134\", "
                                           "\"Google Chrome\";v=\"114.0.5735.134\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-ch-ua-platform-version": "\"15.0.0\"",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "viewport-width": "1475",
            "x-asbd-id": "129477",
            "x-csrftoken": self.csrf_token,
            "x-ig-app-id": self.insta_app_id,
            "x-ig-www-claim": self.x_ig_www_claim,
            "x-requested-with": "XMLHttpRequest",
            "Referer": f"https://www.instagram.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin"
        }

        try:
            http_response = self.request_session.get(
                f"https://www.instagram.com/api/v1/media/{post_id}/likers/",
                headers=request_headers
            )

            response_json: dict = http_response.json()

            if "status" not in response_json or "users" not in response_json:
                return None

            if response_json["status"] != "ok":
                return None

            likers_list = []
            for user in response_json["users"]:
                user: dict
                liker = Liker(
                    user_id=user.get("pk", ""),
                    username=user.get("username", ""),
                    full_name=user.get("full_name", ""),
                    is_private=user.get("is_private", False),
                    badges=user.get("account_badges", []),
                    is_verified=user.get("is_verified", False),
                    profile_picture_id=user.get("profile_pic_id", ""),
                    profile_picture_url=user.get("pprofile_pic_url", ""),
                    latest_reel_media=user.get("latest_reel_media", 0)
                )
                likers_list.append(liker)

            return Likers(
                user_count=response_json.get("user_count", 0),
                users=likers_list
            )
        except JSONDecodeError:
            return None

    def direct(self) -> Direct:
        """
        Lets you use the direct messaging feature

        :return: Direct() class with all the supported methods
        """

        return Direct(self.session_data)



import time
import json
import random
import hashlib
import requests
from uuid import uuid4
from requests import Session
from .lib.Exceptions import NetworkError
from .containers.Inbox import Inbox
from .containers.DirectThread import DirectThread
from .containers.DirectThreadInviter import DirectThreadInviter
from json import JSONDecodeError


class Direct:

    session_data: str
    request_session: Session
    insta_app_id: str = "936619743392459"
    user_id: str
    private_user_agent: str = "Instagram 269.0.0.18.75 Android (26/8.0.0; 480dpi; 1080x1920; " \
                              "OnePlus; 6T Dev; devitron; qcom; en_US; 314665256)"

    def __init__(
        self,
        session_data: str,
        proxy: dict[str, str] = None
    ) -> None:

        self.session_data = session_data
        self.request_session = requests.Session()

        if proxy is not None: self.request_session.proxies.update(proxy)

        session_data_json: dict = json.loads(session_data)
        self.user_id = session_data_json.get("user_id")

        self.request_session.cookies.set("sessionid", session_data_json.get("session_id"))
        self.request_session.cookies.set("rur", session_data_json.get("rur"))
        self.request_session.cookies.set("mid", session_data_json.get("mid"))
        self.request_session.cookies.set("ds_user_id", session_data_json.get("user_id"))
        self.request_session.cookies.set("ig_did", session_data_json.get("ig_did"))

    @property
    def __private_headers(self) -> dict:

        try: mid: str = json.loads(self.session_data).get("mid", "")
        except JSONDecodeError: mid: str = ""

        return {
            "X-IG-App-Locale": "en_US",
            "X-IG-Device-Locale": "en_US",
            "X-IG-Mapped-Locale": "en_US",
            "X-Pigeon-Session-Id": f"UFS-{uuid4()}-1",
            "X-Pigeon-Rawclienttime": str(round(time.time(), 3)),
            "X-IG-Bandwidth-Speed-KBPS": str(random.randint(2500000, 3000000) / 1000),
            "X-IG-Bandwidth-TotalBytes-B": str(random.randint(5000000, 90000000)),
            "X-IG-Bandwidth-TotalTime-MS": str(random.randint(2000, 9000)),
            "X-IG-App-Startup-Country": "IN",
            "X-Bloks-Version-Id": "ce555e5500576acd8e84a66018f54a05720f2dce29f0bb5a1f97f0c10d6fac48",
            "X-IG-WWW-Claim": "0",
            "X-Bloks-Is-Layout-RTL": "false",
            "X-Bloks-Is-Panorama-Enabled": "true",
            "X-IG-Device-ID": str(uuid4()),
            "X-IG-Family-Device-ID": str(uuid4()),
            "X-IG-Android-ID": f"android-{hashlib.sha256(str(time.time()).encode()).hexdigest()[:16]}",
            "X-IG-Timezone-Offset": "-14400",
            "X-IG-Connection-Type": "WIFI",
            "X-IG-Capabilities": "3brTvx0=",
            "X-IG-App-ID": self.insta_app_id,
            "Priority": "u=3",
            "User-Agent": self.private_user_agent,
            "Accept-Language": "en-US",
            "X-MID": mid,
            "Accept-Encoding": "gzip, deflate",
            "Host": "i.instagram.com",
            "X-FB-HTTP-Engine": "Liger",
            "Connection": "keep-alive",
            "X-FB-Client-IP": "True",
            "X-FB-Server-Cluster": "True",
            "IG-INTENDED-USER-ID": self.user_id,
            "X-IG-Nav-Chain": "9MV:self_profile:2,ProfileMediaTabFragment:self_profile:3,9Xf:self_following:4",
            "X-IG-SALT-IDS": str(random.randint(1061162222, 1061262222))
        }

    def inbox(self) -> Inbox:
        """
        Fetches all the chats from inbox.

        :return: Inbox object with chat items and some other data
        """

        # params: json = {
        #     "visual_message_return_type": "unseen",
        #     "thread_message_limit": "10",
        #     "persistentBadging": "true",
        #     "limit": "20",
        #     "is_prefetching": "false",
        #     "fetch_reason": "manual_refresh"
        # }

        http_response = self.request_session.get(
            "https://i.instagram.com/api/v1/direct_v2/inbox/",
            headers=self.__private_headers
        )

        try:
            response_json: dict = http_response.json()

            if response_json.get("status", "") != "ok": raise NetworkError("Key 'status' not 'ok' in response json.")
            if "inbox" not in response_json: raise NetworkError("Key 'inbox' not in response json.")

            inbox_json: dict = response_json.get("inbox", {})

            return Inbox(
                unseen_count=inbox_json.get("unseen_count"),
                unseen_count_timestamp=inbox_json.get("unseen_count_ts"),
                threads=self.__parse_threads(inbox_json.get("threads", list()))
            )

        except JSONDecodeError:
            raise NetworkError("HTTP Response is not a valid JSON.")

    @staticmethod
    def __parse_inviter(data: dict) -> DirectThreadInviter | None:
        if data is None: return None

        return DirectThreadInviter(
            user_id=data.get("pk"),
            username=data.get("username"),
            full_name=data.get("full_name"),
            profile_picture_url=data.get("profile_pic_url"),
            profile_picture_id=data.get("profile_pic_id"),
            is_private=data.get("is_private"),
            is_verified=data.get("is_verified"),
            allowed_commenter_type=data.get("allowed_commenter_type"),
            reel_auto_archive=data.get("reel_auto_archive"),
            has_onboarded_to_text_post_app=data.get("has_onboarded_to_text_post_app"),
            third_party_downloads_enabled=data.get("third_party_downloads_enabled"),
            has_anonymous_profile_picture=data.get("has_anonymous_profile_picture"),
            all_media_count=data.get("all_media_count"),
            liked_clips_count=data.get("liked_clips_count"),
            reachability_status=data.get("reachability_status"),
            has_encrypted_backup=data.get("has_encrypted_backup")
        )

    # @staticmethod
    # def __parse_last_permanent_item(data: dict) -> DirectThreadLastPermanentItem | None:
    #     if data is None: return None
    #
    #     return DirectThreadLastPermanentItem()

    def __parse_threads(self, data: list[dict]) -> list[DirectThread]:

        if len(data) < 1: return list()

        threads: list[DirectThread] = list()

        for item in data:
            theme_data: dict = item.get("theme_data")

            node = DirectThread(
                raw=item,
                thread_title=item.get("thread_title"),
                inviter=self.__parse_inviter(item.get("inviter")),
                # last_permanent_item=self.__parse_last_permanent_item(item.get("last_permanent_item")),
                has_older=item.get("has_older"),
                has_newer=item.get("has_newer"),
                pending=item.get("pending"),
                canonical=item.get("canonical"),
                thread_id=item.get("thread_id"),
                thread_v2_id=item.get("thread_v2_id"),
                viewer_id=item.get("viewer_id"),
                last_activity_at=item.get("last_activity_at"),
                muted=item.get("muted"),
                vc_muted=item.get("vc_muted"),
                approval_required_for_new_members=item.get("approval_required_for_new_members"),
                archived=item.get("archived"),
                thread_has_audio_only_call=item.get("thread_has_audio_only_call"),
                is_group=item.get("is_group"),
                is_translation_enabled=item.get("is_translation_enabled"),
                folder=item.get("folder"),
                e2ee_cutover_status=item.get("e2ee_cutover_status"),
                last_non_sender_item_at=item.get("last_non_sender_item_at"),
                marked_as_unread=item.get("marked_as_unread"),
                input_mode=item.get("input_mode"),
                assigned_admin_id=item.get("assigned_admin_id"),
                mentions_muted=item.get("mentions_muted"),
                is_appointment_booking_enabled=item.get("is_appointment_booking_enabled"),
                is_creator_subscriber_thread=item.get("is_creator_subscriber_thread"),
                business_thread_folder=item.get("business_thread_folder"),
                read_state=item.get("read_state"),
                translation_banner_impression_count=item.get("translation_banner_impression_count"),
                thread_subtype=item.get("thread_subtype"),
                thread_type=item.get("thread_type"),
                is_xac_thread=item.get("is_xac_thread"),
                named=item.get("named"),
                bc_partnership=item.get("bc_partnership"),
                relevancy_score=item.get("relevancy_score"),
                relevancy_score_expr=item.get("relevancy_score_expr"),
                oldest_cursor=item.get("oldest_cursor"),
                newest_cursor=item.get("newest_cursor"),
                next_cursor=item.get("next_cursor"),
                previous_cursor=item.get("prev_cursor"),
                thread_has_drop_in=item.get("thread_has_drop_in"),
                shh_transport_mode=item.get("shh_transport_mode"),
                shh_mode_enabled=item.get("shh_mode_enabled"),
                shh_toggler_userid=item.get("shh_toggler_userid"),
                is_close_friend_thread=item.get("is_close_friend_thread"),
                has_groups_xac_ineligible_user=item.get("has_groups_xac_ineligible_user"),
                system_folder=item.get("system_folder"),
                is_fanclub_subscriber_thread=item.get("is_fanclub_subscriber_thread"),
                joinable_group_link=item.get("joinable_group_link"),
                group_link_joinable_mode=item.get("group_link_joinable_mode"),
                has_reached_message_request_limit=item.get("has_reached_message_request_limit"),
                should_upsell_nudge=item.get("should_upsell_nudge"),
                is_creator_thread=item.get("is_creator_thread"),
                is_business_thread=item.get("is_business_thread"),
                is_sender_possible_business_impersonator=item.get("is_sender_possible_business_impersonator"),
                is_xac_readonly=item.get("is_xac_readonly"),
                will_xac_be_readonly=item.get("will_xac_be_readonly"),
                creator_agent_enabled=item.get("creator_agent_enabled"),
                is_pin=item.get("is_pin"),
                is_spam=item.get("is_spam"),
                spam=item.get("spam"),
                is_3p_api_user=item.get("is_3p_api_user")
            )

            if theme_data is not None:
                node.theme_id = theme_data.get("theme_id")
                node.theme_name = theme_data.get("name")
                node.theme_type = theme_data.get("theme_type")

            threads.append(node)

        return threads



from .PostUser import PostUser
from dataclasses import dataclass, field


@dataclass(frozen=False)
class Post:
    share_url: str = ""
    taken_at: int = 0
    post_id: str = ""
    media_type: int = 0
    code: str = ""
    caption_is_edited: bool = False
    original_media_has_visual_reply_media: bool = False
    like_and_view_counts_disabled: bool = False
    can_viewer_save: bool = False
    profile_grid_control_enabled: bool = False
    is_comments_gif_composer_enabled: bool = False
    comment_threading_enabled: bool = False
    comment_count: int = 0
    has_liked: bool = False
    user: PostUser = field(default_factory=PostUser)
    can_viewer_reshare: bool = False
    like_count: int = 0
    top_likers: list[str] = field(default_factory=list)
    caption_text: str = ""
    is_caption_covered: bool = False
    caption_created_at: int = 0
    caption_share_enabled: bool = False
    caption_did_report_as_spam: bool = False
    is_paid_partnership: bool = False
    show_shop_entrypoint: bool = False
    deleted_reason: int = 0
    integrity_review_decision: str = ""
    ig_media_sharing_disabled: bool = False
    has_shared_to_fb: int = 0
    is_unified_video: bool = False
    should_request_ads: bool = False
    is_visual_reply_commenter_notice_enabled: bool = False
    commerciality_status: str = ""
    explore_hide_comments: bool = False
    has_delayed_metadata: bool = False
    location_latitude: float = 0
    location_longitude: float = 0



from ensta.Guest import Guest
from ensta.SessionHost import SessionHost
from ensta.Host import Host
from ensta.lib import Exceptions
from ensta.PasswordEncryption import PasswordEncryption
from ensta.Authentication import new_session_id
from ensta.Direct import Direct



from dataclasses import dataclass
from .DirectThread import DirectThread


@dataclass(frozen=False)
class Inbox:

    unseen_count: int = None
    unseen_count_timestamp: int = None
    threads: list[DirectThread] = None



from dataclasses import dataclass
from .DirectThreadInviter import DirectThreadInviter
from .DirectThreadLastPermanentItem import DirectThreadLastPermanentItem


@dataclass(frozen=False)
class DirectThread:

    raw: dict = None
    thread_title: str = None
    inviter: DirectThreadInviter = None
    # last_permanent_item: DirectThreadLastPermanentItem = None
    has_older: bool = None
    has_newer: bool = None
    pending: bool = None
    canonical: bool = None
    thread_id: str = None
    thread_v2_id: str = None
    viewer_id: str = None
    last_activity_at: int = None
    muted: bool = None
    vc_muted: bool = None
    approval_required_for_new_members: bool = None
    archived: bool = None
    thread_has_audio_only_call: bool = None
    is_group: bool = None
    is_translation_enabled: bool = None
    folder: int = None
    e2ee_cutover_status: int = None
    last_non_sender_item_at: int = None
    marked_as_unread: bool = None
    input_mode: int = None
    assigned_admin_id: str = None
    mentions_muted: bool = None
    is_appointment_booking_enabled: bool = None
    is_creator_subscriber_thread: bool = None
    business_thread_folder: int = None
    read_state: int = None
    translation_banner_impression_count: int = None
    thread_subtype: int = None
    thread_type: str = None
    is_xac_thread: bool = None
    named: bool = None
    bc_partnership: bool = None
    relevancy_score: int = None
    relevancy_score_expr: int = None
    oldest_cursor: str = None
    newest_cursor: str = None
    next_cursor: str = None
    previous_cursor: str = None
    thread_has_drop_in: bool = None
    shh_transport_mode: int = None
    shh_mode_enabled: bool = None
    shh_toggler_userid: int = None
    theme_id: str = None
    theme_type: int = None
    theme_name: str = None
    is_close_friend_thread: bool = None
    has_groups_xac_ineligible_user: bool = None
    system_folder: int = None
    is_fanclub_subscriber_thread: bool = None
    joinable_group_link: str = None
    group_link_joinable_mode: int = None
    has_reached_message_request_limit: bool = None
    should_upsell_nudge: bool = None
    is_creator_thread: bool = None
    is_business_thread: bool = None
    is_sender_possible_business_impersonator: bool = None
    is_xac_readonly: bool = None
    will_xac_be_readonly: bool = None
    creator_agent_enabled: bool = None
    is_pin: bool = None
    is_spam: bool = None
    spam: bool = None
    is_3p_api_user: bool = None



import json
import random
import string
import requests
from requests import Session
from requests import Response
from json import JSONDecodeError
from .PasswordEncryption import PasswordEncryption
from .lib.Exceptions import (AuthenticationError, NetworkError)


def new_session_id(username: str, password: str, proxy: dict[str, str] | None = None) -> str:
    request_session: Session = requests.Session()

    if proxy is not None: request_session.proxies.update(proxy)

    encryption = PasswordEncryption(request_session)
    encrypted_password = encryption.encrypt(password)

    data: dict = {
        "enc_password": encrypted_password,
        "optIntoOneTap": False,
        "queryParams": "{}",
        "trustedDeviceRecords": "{}",
        "username": username
    }

    csrf_token: str = "".join(random.choices(string.ascii_letters + string.digits, k=32))

    headers: dict = {
        "accept": "*/*",
        "accept-language": "en-US,en;q=0.9",
        "content-type": "application/x-www-form-urlencoded",
        "dpr": "1.30208",
        "sec-ch-prefers-color-scheme": "dark",
        "sec-ch-ua": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                     "Chrome/119.0.0.0 Safari/537.36",
        "sec-ch-ua-full-version-list": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                                       "(KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-model": "\"\"",
        "sec-ch-ua-platform": "\"Windows\"",
        "sec-ch-ua-platform-version": "\"15.0.0\"",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "viewport-width": "1475",
        "x-asbd-id": "129477",
        "x-csrftoken": csrf_token,
        "x-ig-app-id": "936619743392459",
        "x-ig-www-claim": "0",
        "x-instagram-ajax": "1009977574",
        "x-requested-with": "XMLHttpRequest",
        "x-web-device-id": "25532C62-8BBC-4927-B6C5-02631D6E05BF",
        "cookie": f"dpr=1.3020833730697632; csrftoken={csrf_token}",
        "Referer": "https://www.instagram.com/",
        "Referrer-Policy": "strict-origin-when-cross-origin"
    }

    http_response: Response = request_session.post(
        "https://www.instagram.com/api/v1/web/accounts/login/ajax/",
        data=data,
        headers=headers
    )

    try:
        response_json: dict = http_response.json()

        if response_json.get("status", "") != "ok": raise AuthenticationError("User doesn't exist.")

        if response_json.get("user", False) is False or response_json.get("authenticated", False) is False:
            raise AuthenticationError("Invalid password.")

        session_id: str = http_response.cookies.get("sessionid", "")
        rur: str = http_response.cookies.get("rur", "")
        mid: str = http_response.cookies.get("mid", "")
        user_id: str = response_json.get("userId", "")
        ig_did: str = response_json.get("ig_did", "")

        if session_id == "" or user_id == "": raise AuthenticationError("Unable to login.")

        return json.dumps({
            "session_id": session_id,
            "rur": rur,
            "mid": mid,
            "user_id": user_id,
            "ig_did": ig_did,
            "username": username
        })

    except JSONDecodeError:
        raise NetworkError("Response got while logging in was not a valid "
                           "json. Are you able to visit Instagram on the web?")



class SessionError(Exception):

    def __init__(self, message):
        super().__init__(message)


class NetworkError(Exception):

    def __init__(self, message):
        super().__init__(message)


class IdentifierError(Exception):

    def __init__(self, message):
        super().__init__(message)


class DevelopmentError(Exception):

    def __init__(self):
        super().__init__(
            f"There was an error while executing this function, maybe caused due to a bug in the "
            "code. Please raise an issue on GitHub"
        )


class AuthenticationError(Exception):

    def __init__(self, message):
        super().__init__(message)


class APIError(Exception):

    def __init__(self, message: str | None = None):
        if message is not None:
            super().__init__(
                "Looks like Instagram has made changes to the API Response."
                "Please raise this as an issue on GitHub."
            )
        else:
            super().__init__(message)


class ConversionError(Exception):

    def __init__(self, message):
        super().__init__(message)


class FileTypeError(Exception):

    def __init__(self, message):
        super().__init__(message)



import time
import base64
from requests import Session
from Cryptodome.PublicKey import RSA
from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.Random import get_random_bytes


class PasswordEncryption:

    request_session: Session = None

    def __init__(self, request_session: Session):
        self.request_session = request_session

    def encrypt(self, password: str) -> str:
        public_key, public_key_id = self.__public_keys()
        session_key = get_random_bytes(32)
        iv = get_random_bytes(12)
        timestamp = str(int(time.time()))
        decoded_public_key = base64.b64decode(public_key.encode())
        recipient_key = RSA.importKey(decoded_public_key)
        cipher_rsa = PKCS1_v1_5.new(recipient_key)
        rsa_encrypted = cipher_rsa.encrypt(session_key)
        cipher_aes = AES.new(session_key, AES.MODE_GCM, iv)
        cipher_aes.update(timestamp.encode())
        aes_encrypted, tag = cipher_aes.encrypt_and_digest(password.encode("utf8"))
        size_buffer = len(rsa_encrypted).to_bytes(2, byteorder="little")

        payload = base64.b64encode(b"".join([
            b"\x01",
            public_key_id.to_bytes(1, byteorder="big"),
            iv,
            size_buffer,
            rsa_encrypted,
            tag,
            aes_encrypted
        ]))

        return f"#PWD_INSTAGRAM:4:{timestamp}:{payload.decode()}"

    def __public_keys(self) -> tuple[str, int]:
        response = self.request_session.get("https://i.instagram.com/api/v1/qe/sync/")

        public_key: str = response.headers.get("ig-set-password-encryption-pub-key")
        public_key_id: int = int(response.headers.get("ig-set-password-encryption-key-id"))

        return public_key, public_key_id



from distutils.core import setup
from pathlib import Path

version = "5.0.3.8"
long_description = (Path(__file__).parent / "README.md").read_text(encoding="utf-8")

setup(
    name="ensta",
    packages=["ensta", "ensta.lib", "ensta.containers"],
    version=version,
    license="MIT",
    description="🔥 Fastest & Simplest Python Package For Instagram Automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Deepak Soni",
    author_email="sonniiii@outlook.com",
    url="https://github.com/diezo/ensta",
    download_url=f"https://github.com/diezo/ensta/archive/refs/tags/v{version}.tar.gz",
    keywords=[
        "instagram-client",
        "instagram",
        "api-wrapper",
        "instagram-scraper",
        "instagram-api",
        "instagram-sdk",
        "instagram-photos",
        "instagram-api-python",
        "instabot",
        "instagram-stories",
        "instagram-bot",
        "instapy",
        "instagram-downloader",
        "instagram-account",
        "instagram-crawler",
        "instagram-private-api",
        "igtv",
        "instagram-automation",
        "reels",
        "instagram-feed"
    ],
    install_requires=[
        "requests",
        "moviepy",
        "pycryptodome"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10"
    ]
)



from dataclasses import dataclass


@dataclass(frozen=False)
class DirectThreadInviter:

    user_id: str = None
    username: str = None
    full_name: str = None
    profile_picture_url: str = None
    profile_picture_id: str = None
    is_private: bool = None
    is_verified: bool = None
    allowed_commenter_type: str = None
    reel_auto_archive: str = None
    has_onboarded_to_text_post_app: bool = None
    third_party_downloads_enabled: int = None
    has_anonymous_profile_picture: bool = None
    all_media_count: int = None
    liked_clips_count: int = None
    reachability_status: int = None
    has_encrypted_backup: bool = None
